from iminuit import Minuit
import warnings
import numpy as np


def minimize(fun, x0, args=(), method=None,
             jac=None, hess=None, hessp=None,
             bounds=None, constraints=None,
             tol=None, callback=None, options=None):
    """An interface to MIGRAD using the ``scipy.optimize.minimize`` API.

    For a general description of the arguments, see ``scipy.optimize.minimize``.

    The ``method`` argument is ignored. The optimisation is always done using MIGRAD.

    The `options` argument can be used to pass special settings to Minuit.
    All are optional.

    **Options:**

      - *disp* (bool): Set to true to print convergence messages. Default: False.
      - *maxfev* (int): Maximum allowed number of iterations. Default: 10000.
      - *eps* (sequence): Initial step size to numerical compute derivative.
        Minuit automatically refines this in subsequent iterations and is very
        insensitive to the initial choice. Default: 1.

    **Returns: OptimizeResult** (dict with attribute access)
      - *x* (ndarray): Solution of optimization.
      - *fun* (float): Value of objective function at minimum.
      - *message* (str): Description of cause of termination.
      - *hess_inv* (ndarray): Inverse of Hesse matrix at minimum (may not be exact).
      - nfev (int): Number of function evaluations.
      - njev (int): Number of jacobian evaluations.
      - minuit (object): Minuit object internally used to do the minimization.
        Use this to extract more information about the parameter errors.
    """
    from scipy.optimize import OptimizeResult

    if method not in {None, 'migrad'}:
        warnings.warn("method argument is ignored")

    if constraints is not None:
        raise ValueError("Constraints are not supported by Migrad, only bounds")

    if hess or hessp:
        warnings.warn("hess and hessp arguments cannot be handled and are ignored")

    if constraints:
        raise ValueError("Minuit only supports bounds, not constraints")

    if tol:
        warnings.warn("tol argument has no effect on Minuit")

    def wrapped(func, args, callback=None):
        if callback is None:
            return lambda x: func(x, *args)
        else:
            def f(x):
                callback(x)
                return func(x, *args)
            return f
    wrapped_fun = wrapped(fun, args, callback)

    maxfev = 10000
    error = None
    kwargs = {'print_level': 0, 'errordef': 1}
    if options:
        if "disp" in options:
            kwargs['print_level'] = 1
        if "maxiter" in options:
            warnings.warn("maxiter not supported, acts like maxfev instead")
            maxfev = options["maxiter"]
        if "maxfev" in options:
            maxfev = options["maxfev"]
        if "eps" in options:
            error = options["eps"]

    # prevent warnings from Minuit about missing initial step
    if error is None:
        error = np.ones_like(x0)

    if bool(jac):
        if jac is True:
            raise ValueError("jac=True is not supported, only jac=callable")
        assert hasattr(jac, "__call__")
        wrapped_grad = wrapped(jac, args)
    else:
        wrapped_grad = None

    m = Minuit.from_array_func(wrapped_fun,
                               x0,
                               error=error,
                               limit=bounds,
                               grad=wrapped_grad,
                               **kwargs)
    m.migrad(ncall=maxfev)

    message = "Optimization terminated successfully."
    success = m.migrad_ok()
    if not success:
        message = "Optimization failed."
        fmin = m.get_fmin()
        if fmin.has_reached_call_limit:
            message += " Call limit was reached."
        if fmin.is_above_max_edm:
            message += " Estimated distance to minimum too large."

    return OptimizeResult(
        x=m.np_values(),
        success=success,
        fun=m.fval,
        hess_inv=m.np_covariance(),
        message=message,
        nfev=m.get_num_call_fcn(),
        njev=m.get_num_call_grad(),
        minuit=m,
    )



"""MINUIT from Python - Fitting like a boss

Basic usage example::

    from iminuit import Minuit
    def f(x, y, z):
        return (x - 2) ** 2 + (y - 3) ** 2 + (z - 4) ** 2
    m = Minuit(f)
    m.migrad()
    print(m.values)  # {'x': 2,'y': 3,'z': 4}
    print(m.errors)  # {'x': 1,'y': 1,'z': 1}

Further information:

* Code: https://github.com/iminuit/iminuit
* Docs: https://iminuit.readthedocs.io
"""

__all__ = [
    'Minuit',
    'minimize',
    'describe',
    'Struct',
    '__version__',
    'test',
]

from ._libiminuit import Minuit
from ._minimize import minimize
from .util import describe, Struct
from .info import __version__


def test(args=None):
    """Execute the iminuit tests.

    Requires pytest.

    From the command line:

        python -c 'import iminuit; iminuit.test()
    """
    # http://pytest.org/latest/usage.html#calling-pytest-from-python-code
    import pytest
    args = ['-v', '--pyargs', 'iminuit']
    pytest.main(args)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import sys
from .frontend import Frontend
from math import log10

__all__ = ['ConsoleFrontend']


class ConsoleFrontend(Frontend):
    """Console frontend for Minuit.

    This class prints stuff directly via print.
    """

    def display(self, *args):
        sys.stdout.write('\n'.join(args) + '\n')

    def print_fmin(self, fmin, tolerance=None, ncalls=0):
        """display function minimum information
        for FunctionMinimumStruct *sfmin*.
        It contains various migrad status.
        """
        goaledm = 0.0001 * tolerance * fmin.up if tolerance is not None else ''
        # despite what the doc said the code is actually 1e-4
        # http://wwwasdoc.web.cern.ch/wwwasdoc/hbook_html3/node125.html
        flatlocal = dict(list(locals().items()) + list(fmin.items()))
        info1 = 'fval = %(fval)r | total call = %(ncalls)r | ncalls = %(nfcn)r' % \
                flatlocal
        info2 = 'edm = %(edm)s (Goal: %(goaledm)s) | up = %(up)r' % flatlocal
        header1 = '|' + (' %14s |' * 5) % (
            'Valid',
            'Valid Param',
            'Accurate Covar',
            'Posdef',
            'Made Posdef')
        hline = '-' * len(header1)
        status1 = '|' + (' %14r |' * 5) % (
            fmin.is_valid,
            fmin.has_valid_parameters,
            fmin.has_accurate_covar,
            fmin.has_posdef_covar,
            fmin.has_made_posdef_covar)
        header2 = '|' + (' %14s |' * 5) % (
            'Hesse Fail',
            'Has Cov',
            'Above EDM',
            '',
            'Reach calllim')
        status2 = '|' + (' %14s |' * 5) % (
            fmin.hesse_failed,
            fmin.has_covariance,
            fmin.is_above_max_edm,
            '',
            fmin.has_reached_call_limit)

        self.display(hline, info1, info2,
                     hline, header1, hline, status1,
                     hline, header2, hline, status2,
                     hline)

    def print_merror(self, vname, smerr):
        """print minos error for varname"""
        stat = 'VALID' if smerr.is_valid else 'PROBLEM'

        summary = 'Minos Status for %s: %s' % \
                  (vname, stat)

        error = '| {0:^15s} | {1:^12g} | {2:^12g} |'.format(
            'Error',
            smerr.
                lower,
            smerr.upper)
        valid = '| {0:^15s} | {1:^12s} | {2:^12s} |'.format(
            'Valid',
            str(smerr.lower_valid),
            str(smerr.upper_valid))
        at_limit = '| {0:^15s} | {1:^12s} | {2:^12s} |'.format(
            'At Limit',
            str(smerr.at_lower_limit),
            str(smerr.at_upper_limit))
        max_fcn = '| {0:^15s} | {1:^12s} | {2:^12s} |'.format(
            'Max FCN',
            str(smerr.at_lower_max_fcn),
            str(smerr.at_upper_max_fcn))
        new_min = '| {0:^15s} | {1:^12s} | {2:^12s} |'.format(
            'New Min',
            str(smerr.lower_new_min),
            str(smerr.upper_new_min))
        hline = '-' * len(error)
        self.display(hline, summary, hline, error, valid,
                     at_limit, max_fcn, new_min, hline)

    def print_param(self, mps, merr=None, float_format=None):
        """Print parameter states

        Arguments:

            *mps*: list of MinuitParameter struct

            *merr*: dictionary of vname->minos error struct

            *float_format*: ignored
        """
        merr = {} if merr is None else merr
        vnames = [mp.name for mp in mps]
        name_width = max([len(x) for x in vnames]) if vnames else 0
        name_width = max(4, name_width)
        num_max = len(vnames) - 1
        num_width = max(2, int(log10(max(num_max, 1)) + 1))

        header = (('| {0:^%is} | {1:^%is} | {2:^8s} | {3:^8s} | {4:^8s} |'
                   ' {5:^8s} | {6:8s} | {7:8s} | {8:^6s} |') %
                  (num_width, name_width)).format(
            'No', 'Name', 'Value', 'Sym. Err',
            "Err-", "Err+", "Limit-", "Limit+", "Fixed?")
        hline = '-' * len(header)
        linefmt = (('| {0:>%id} | {1:>%is} | {2:<8s} | {3:<8s} | {4:<8s} |'
                    ' {5:<8s} | {6:8s} | {7:8s} | {8:^6s} |') %
                   (num_width, name_width))
        nfmt = '{0:<8.4G}'
        nformat = nfmt.format
        blank = ' ' * 8

        tab = [hline, header, hline]
        for i, (v, mp) in enumerate(zip(vnames, mps)):
            tmp = [i, v]

            tmp.append(nfmt.format(mp.value))
            tmp.append(nfmt.format(mp.error))

            tmp.append(nformat(merr[v].lower) if v in merr else blank)
            tmp.append(nformat(merr[v].upper) if v in merr else blank)

            tmp.append(nformat(mp.lower_limit) if mp.lower_limit is not None else blank)
            tmp.append(nformat(mp.upper_limit) if mp.upper_limit is not None else blank)

            tmp.append(
                'Yes' if mp.is_fixed else 'CONST' if mp.is_const else 'No')

            line = linefmt.format(*tmp)
            tab.append(line)
        tab.append(hline)
        self.display(*tab)

    def print_banner(self, cmd):
        """show banner of command"""
        hline = '*' * 50
        migrad = '*{:^48}*'.format(cmd)
        self.display(hline, migrad, hline + '\n')

    def print_matrix(self, vnames, matrix):
        def row_fmt(args):
            s = '| ' + args[0] + ' |'
            for x in args[1:]:
                s += ' ' + x
            s += ' |'
            return s

        row_width = max(max(len(v) for v in vnames), 4)
        blank = ' ' * row_width
        vnames = [('{:>%is}' % row_width).format(x) for x in vnames]
        val_fmt = '{:%i.2f}' % row_width

        header = row_fmt([blank] + vnames)
        hline = '-' * len(header)
        tab = [hline, "Correlation", hline, header, hline]
        for (vn, row) in zip(vnames, matrix):
            tab.append(row_fmt([vn] + [val_fmt.format(x) for x in row]))
        tab.append(hline)
        self.display(*tab)

    def print_hline(self, width=86):
        self.display('*' * width)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from iminuit.util import (fitarg_rename,
                          true_param,
                          param_name,
                          extract_iv,
                          extract_limit,
                          extract_error,
                          extract_fix,
                          remove_var,
                          arguments_from_docstring,
                          Struct)
import pytest


def test_fitarg_rename():
    fitarg = {'x': 1, 'limit_x': (2, 3), 'fix_x': True, 'error_x': 10}

    def ren(x):
        return 'z_' + x

    newfa = fitarg_rename(fitarg, ren)
    assert 'z_x' in newfa
    assert 'limit_z_x' in newfa
    assert 'error_z_x' in newfa
    assert 'fix_z_x' in newfa
    assert len(newfa) == 4


def test_fitarg_rename_strprefix():
    fitarg = {'x': 1, 'limit_x': (2, 3), 'fix_x': True, 'error_x': 10}
    newfa = fitarg_rename(fitarg, 'z')
    assert 'z_x' in newfa
    assert 'limit_z_x' in newfa
    assert 'error_z_x' in newfa
    assert 'fix_z_x' in newfa
    assert len(newfa) == 4


def test_true_param():
    assert true_param('N') is True
    assert true_param('limit_N') is False
    assert true_param('error_N') is False
    assert true_param('fix_N') is False


def test_param_name():
    assert param_name('N') == 'N'
    assert param_name('limit_N') == 'N'
    assert param_name('error_N') == 'N'
    assert param_name('fix_N') == 'N'


def test_extract_iv():
    d = dict(k=1., limit_k=1., error_k=1., fix_k=1.)
    ret = extract_iv(d)
    assert 'k' in ret
    assert 'limit_k' not in ret
    assert 'error_k' not in ret
    assert 'fix_k' not in ret


def test_extract_limit():
    d = dict(k=1., limit_k=1., error_k=1., fix_k=1.)
    ret = extract_limit(d)
    assert 'k' not in ret
    assert 'limit_k' in ret
    assert 'error_k' not in ret
    assert 'fix_k' not in ret


def test_extract_error():
    d = dict(k=1., limit_k=1., error_k=1., fix_k=1.)
    ret = extract_error(d)
    assert 'k' not in ret
    assert 'limit_k' not in ret
    assert 'error_k' in ret
    assert 'fix_k' not in ret


def test_extract_fix():
    d = dict(k=1., limit_k=1., error_k=1., fix_k=1.)
    ret = extract_fix(d)
    assert 'k' not in ret
    assert 'limit_k' not in ret
    assert 'error_k' not in ret
    assert 'fix_k' in ret


def test_remove_var():
    dk = dict(k=1, limit_k=1, error_k=1, fix_k=1)
    dl = dict(l=1, limit_l=1, error_l=1, fix_l=1)
    dm = dict(m=1, limit_m=1, error_m=1, fix_m=1)
    dn = dict(n=1, limit_n=1, error_n=1, fix_n=1)
    d = {}
    d.update(dk)
    d.update(dl)
    d.update(dm)
    d.update(dn)

    ret = remove_var(d, ['k', 'm'])
    for k in dk:
        assert k not in ret
    for k in dl:
        assert k in ret
    for k in dm:
        assert k not in ret
    for k in dn:
        assert k in ret


def test_arguments_from_docstring():
    s = 'f(x, y, z)'
    a = arguments_from_docstring(s)
    assert a == ['x', 'y', 'z']
    # this is a hard one
    s = 'Minuit.migrad( int ncall_me =10000, [resume=True, int nsplit=1])'
    a = arguments_from_docstring(s)
    assert a == ['ncall_me', 'resume', 'nsplit']


def test_Struct():
    s = Struct(a=1, b=2)
    assert set(s.keys()) == {'a', 'b'}
    assert s.a == 1
    assert s.b == 2
    s.a = 3
    assert s.a == 3
    with pytest.raises(AttributeError):
        s.c
    with pytest.raises(KeyError):
        s['c']
    assert s == Struct(a=3, b=2)
